import asyncio
import base64
import datetime
import itertools
from collections import defaultdict, namedtuple
from io import BytesIO
from hashlib import md5

from aio_pyformance.registry import MetricsRegistry
import logging
import re
import rapidjson as json
from botocore.exceptions import ClientError
from aiohttp.web_exceptions import HTTPBadRequest

from feed_manager.api.models.internal_models import ExtendedReport
from pscr_common.utils import (
    chunker,
    get_utcnow_with_tzinfo,
    datetime_to_sec_timestamp,
    datetime_to_ms_timestamp,
    EPOCH,
)
from sqlalchemy import and_, or_, tuple_, exists, null, true, func
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy import bindparam, delete, select
from sqlalchemy.exc import IntegrityError
from typing import List, Dict, Optional, Callable, NamedTuple, Iterable, Collection
import uuid

from feed_manager.api.models import (
    Report as ApiReport,
    IOC_V2,
    AugmentedWatchlist,
    WatchlistAction,
    AugmentedWatchlists,
    WatchlistMeta,
    AugmentedReport,
    ReportSeverity,
    WatchlistV2,
    ClassifierKeyValue,
    ReportIOCIgnore,
    WatchlistTelemetry,
)
from feed_manager.engine.errors import EntityNotFound
from feed_manager.engine.org_watchlist_state_updater import OrgWatchlistStateUpdater
from feed_manager.engine.report_engine import ReportEngine, ReportNotFound, IOCNotFound
from feed_manager.engine.feed_engine import FeedEngine
from feed_manager.sql.models import (
    Report as DbReport,
    IOC,
    Watchlist,
    IgnoreReport,
    IgnoreReportIOC,
    ReportSeverityOverride,
    Classifier,
    ClassifierWatchlist as DbClassifierWatchlist,
    ReportWatchlist as DbReportWatchlist,
    ReportClassifier,
    OrganizationWatchlistState,
    OrganizationWatchlistBuild,
    Feed as DbFeed,
)
from feed_manager.utils import json_dumps, aws_client
from feed_manager.utils.config import Config
from feed_manager.utils.db import (
    ReadWriteDbPool,
    is_unique_violation_exception,
    is_fk_violation_exception,
)
from feed_manager.utils.mem_cache import MemCache as Cache
from feed_manager.utils.publisher import Publisher
from feed_manager.utils.telemetry_client import TelemetryClient

_logger = logging.getLogger(__name__)


class WatchlistNotFound(EntityNotFound):
    def __init__(self, watchlist_id):
        super(WatchlistNotFound, self).__init__(f"Watchlist '{watchlist_id}' not found")


class InvalidWatchlistAlerts(HTTPBadRequest):
    def __init__(self):
        super(InvalidWatchlistAlerts, self).__init__(
            reason="Alert generation is not supported without tagging enabled. "
            "Consult the Developer Network documentation for this "
            "endpoint for more information"
        )


class DirtyOrgAugmentationTuple(NamedTuple):
    org_id: str
    dirty_timestamp: datetime.datetime
    rebuild_timestamp: datetime.datetime

    def queue_time_sec(self):
        return (self.rebuild_timestamp - self.dirty_timestamp).total_seconds()


class WatchlistEngine(object):
    VALID_ID_RE = re.compile(r"^[\w-]+$", re.ASCII)

    def __init__(
        self,
        config: Config,
        db_pool: ReadWriteDbPool,
        publisher: Publisher,
        metrics: MetricsRegistry,
        report_engine: ReportEngine,
    ):
        self.db_pool = db_pool
        self.metrics = metrics
        self.publisher = publisher
        self.report_engine = report_engine
        self.config = config
        self.telemetry_client = TelemetryClient(self.config)

    async def shutdown(self, _app=None):
        await self.telemetry_client.close()
        await self.report_engine.shutdown()

    async def _check_is_alerting_forbidden(self, watchlist: WatchlistV2) -> bool:
        if (
            watchlist.classifier
            and watchlist.classifier.key == FeedEngine.FEED_ID_CLASSIFIER
        ):
            async with self.db_pool.ro_transaction() as conn:
                sel_stmt = select([DbFeed]).where(
                    DbFeed.id == watchlist.classifier.value
                )
                feeds = []
                async for result in await conn.stream(sel_stmt):
                    feeds.append(result[0])
                if len(feeds) and not feeds[0].alertable:
                    return True
        return False

    async def _is_alert_classification_feature_enabled(
        self, watchlist: WatchlistV2
    ) -> bool:
        if (
            watchlist.classifier
            and watchlist.classifier.key == FeedEngine.FEED_ID_CLASSIFIER
        ):
            async with self.db_pool.ro_transaction() as conn:
                sel_stmt = select([DbFeed]).where(
                    DbFeed.id == watchlist.classifier.value
                )
                feeds = []
                async for result in await conn.stream(sel_stmt):
                    feeds.append(result[0])

                if (
                    len(feeds)
                    and feeds[0].alert_classification_feature_enabled is not None
                    and feeds[0].alert_classification_feature_enabled
                ):
                    return True
        return False

    @staticmethod
    def gen_watchlist_id():
        return (
            base64.b64encode(uuid.uuid4().bytes)
            .decode("utf-8")
            .rstrip("=\n")
            .replace("/", "")
            .replace("+", "")
        )

    async def _update_dirty_watchlist_org(
        self, conn, org_id: str, now: datetime.datetime = None
    ) -> None:
        now = now if now else get_utcnow_with_tzinfo()

        await OrgWatchlistStateUpdater(self.config, now).set_dirty_for_org_id(
            org_id
        ).execute(conn)

    async def add_watchlist(
        self, org_id: str, watchlist: WatchlistV2, now: datetime.datetime
    ) -> WatchlistV2:
        tags = watchlist.tags_enabled if watchlist.tags_enabled is not None else True
        alerts = (
            watchlist.alerts_enabled if watchlist.alerts_enabled is not None else False
        )

        if alerts and not tags:
            _logger.warning(
                "Preventing org %s from enabling alerts without tags enabled on new watchlist",
                org_id,
            )
            raise InvalidWatchlistAlerts()

        if alerts and await self._check_is_alerting_forbidden(watchlist):
            _logger.warning(
                "Preventing org %s from enabling alerts on new watchlist", org_id
            )
            raise HTTPBadRequest(
                reason="Alert generation is not supported for this feed. "
                + "Consult the Developer Network documentation for this "
                + "endpoint for more information"
            )

        # Prevent creation of custom watchlist with the same name as a public feed
        public_feed_names = Cache().get("public_feeds")
        if (
            public_feed_names
            and watchlist.name in public_feed_names
            and not watchlist.classifier
        ):
            _logger.warning(
                f"Could not create watchlist with name '{watchlist.name}' for org {org_id} because its "
                f"name conflicts with a public feed."
            )
            raise HTTPBadRequest(
                reason=f"Cannot create a watchlist with name '{watchlist.name}' because a public feed with that name "
                f"already exists."
            )

        async with self.db_pool.rw_transaction() as conn:
            watchlist_id = await self._create_watchlist(
                conn,
                org_id,
                watchlist.id,
                watchlist.name,
                description=watchlist.description,
                tags=tags,
                alerts=alerts,
                create_timestamp=now,
                last_update_timestamp=now,
                alert_classification_enabled=None,
            )

            if watchlist.classifier:
                classifier_id = await self.report_engine.get_classifier_id(
                    watchlist.classifier.key, watchlist.classifier.value, conn=conn
                )

                await conn.execute(
                    DbClassifierWatchlist.__table__.insert().values(
                        watchlist_id=watchlist_id, classifier_id=classifier_id
                    )
                )

            if watchlist.report_ids:
                await self._add_watchlist_report_ids(
                    conn,
                    org_id=org_id,
                    watchlist_id=watchlist_id,
                    report_ids=watchlist.report_ids,
                )

            await self._update_dirty_watchlist_org(conn, org_id, now=now)

        return WatchlistV2(
            id=watchlist_id,
            name=watchlist.name,
            description=watchlist.description,
            classifier=watchlist.classifier if watchlist.classifier else None,
            report_ids=watchlist.report_ids if watchlist.report_ids else None,
            tags_enabled=tags,
            alerts_enabled=alerts,
            create_timestamp=now,
            last_update_timestamp=now,
            alert_classification_enabled=watchlist.alert_classification_enabled,
        )

    @staticmethod
    async def _add_watchlist_report_ids(
        conn, org_id: str, watchlist_id: str, report_ids: Collection[str]
    ) -> None:
        if not report_ids:
            return

        join_stmt = DbReport.__table__.join(
            ReportClassifier.__table__, ReportClassifier.report_id == DbReport.id
        ).join(Classifier.__table__, Classifier.id == ReportClassifier.classifier_id)

        sel_stmt = (
            select([bindparam("watchlist_id", watchlist_id), DbReport.id])
            .select_from(join_stmt)
            .where(
                or_(
                    and_(
                        Classifier.key == ReportEngine.SOURCE_CLASSIFIER,
                        Classifier.value == org_id,
                    ),
                    and_(
                        Classifier.key == ReportEngine.ACCESS_CLASSIFIER,
                        Classifier.value == "public",
                    ),
                )
            )
            .where(DbReport.id.in_(report_ids))
            .group_by(DbReport.id)
        )

        insert_stmt = (
            insert(DbReportWatchlist)
            .returning(DbReportWatchlist.report_id)
            .from_select(
                (DbReportWatchlist.watchlist_id, DbReportWatchlist.report_id), sel_stmt
            )
        )

        # Use do_update here so that we can use return ids.  In reality we don't update anything
        upsert_stmt = insert_stmt.on_conflict_do_update(
            index_elements=[
                DbReportWatchlist.watchlist_id,
                DbReportWatchlist.report_id,
            ],
            set_={"watchlist_id": insert_stmt.excluded.watchlist_id},
        )

        try:
            results = (await conn.execute(upsert_stmt)).fetchall()

            # Make sure all the report ids passed the public/org_id join test above
            if len(results) != len(report_ids):
                diff = set(report_ids) - set(result["report_id"] for result in results)
                raise ValueError(
                    f"Invalid report id(s) in watchlist: {', '.join(diff)}"
                )
        except IntegrityError as e:
            if not is_fk_violation_exception(e):
                raise e

            _logger.info(
                "Unable to update report watchlist.  Watchlist '%s' not found",
                watchlist_id,
            )
            raise WatchlistNotFound(watchlist_id)

    @classmethod
    async def _create_watchlist(
        cls,
        conn,
        org_id: str,
        watchlist_id: str,
        name: str,
        create_timestamp: datetime.datetime,
        last_update_timestamp: datetime.datetime,
        description: str = None,
        tags: bool = False,
        alerts: bool = False,
        alert_classification_enabled: bool = None,
    ) -> str:
        watchlist_id = watchlist_id if watchlist_id else cls.gen_watchlist_id()

        try:
            await conn.execute(
                Watchlist.__table__.insert().values(
                    id=watchlist_id,
                    org_id=org_id,
                    name=name,
                    description=description,
                    tags_enabled=tags,
                    alerts_enabled=alerts,
                    create_timestamp=create_timestamp,
                    last_update_timestamp=last_update_timestamp,
                    alert_classification_enabled=alert_classification_enabled,
                )
            )
        except IntegrityError:
            raise ValueError(f"Watchlist {name} already exists")

        return watchlist_id

    async def remove_watchlist(self, org_id: str, watchlist_id: str) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            del_stmt = (
                delete(Watchlist)
                .where(Watchlist.org_id == org_id)
                .where(Watchlist.id == watchlist_id)
            )

            result = await conn.execute(del_stmt)

            if result.rowcount != 0:
                await self._update_dirty_watchlist_org(conn, org_id=org_id)
                return True
            else:
                return False

    async def remove_all_watchlists(self, org_id: str) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            # If there are no watchlists, don't need to do anything
            sel_stmt = select([func.count(Watchlist.id)]).where(
                Watchlist.org_id == org_id
            )
            select_result = (await conn.execute(sel_stmt)).scalar()

            if select_result == 0:
                return True

            del_stmt = delete(Watchlist).where(Watchlist.org_id == org_id)
            delete_result = await conn.execute(del_stmt)
            if delete_result.rowcount != 0:
                await self._update_dirty_watchlist_org(conn, org_id=org_id)
                return True
            else:
                return False

    async def update_watchlist(
        self, org_id: str, watchlist: WatchlistV2, now: datetime.datetime
    ) -> WatchlistV2:
        tags = watchlist.tags_enabled
        db_watchlists = await self.get_watchlists(org_id, watchlist.id)

        if tags is None:
            # we assume False as default to prevent watchlists ending up in unhealthy state
            tags = db_watchlists[0].tags_enabled if db_watchlists else False

        if watchlist.alerts_enabled and not tags:
            _logger.warning(
                "Preventing org %s from enabling alerts without tags enabled on watchlist %s",
                org_id,
                watchlist.id,
            )
            raise InvalidWatchlistAlerts()

        if watchlist.classifier:
            classifier_id = await self.report_engine.get_classifier_id(
                watchlist.classifier.key, watchlist.classifier.value
            )
        else:
            classifier_id = None

        if (
            await self._check_is_alerting_forbidden(watchlist)
            and watchlist.alerts_enabled
        ):
            _logger.warning(
                "Preventing org %s from enabling alerts on updated watchlist %s",
                org_id,
                watchlist.id,
            )
            raise HTTPBadRequest(
                reason="Alert generation is not supported for this feed. "
                + "Consult the Developer Network documentation for this "
                + "endpoint for more information"
            )

        if watchlist.name and watchlist.name == db_watchlists[0].name:
            perform_name_validation = False
        else:
            perform_name_validation = True

        # Prevent update of custom watchlist to the same name as a public feed
        public_feed_names = Cache().get("public_feeds")
        if (
            perform_name_validation
            and public_feed_names
            and watchlist.name in public_feed_names
            and not watchlist.classifier
        ):
            _logger.warning(
                f"Could not update watchlist to name '{watchlist.name}' for org {org_id} because its "
                f"name conflicts with a public feed."
            )
            raise HTTPBadRequest(
                reason=f"Cannot update a watchlist to name '{watchlist.name}' because a public feed with that name "
                f"already exists."
            )

        # Prevent enabling/ disabling alert classification if the feature is not turned on for this watchlist
        if (
            watchlist.alert_classification_enabled
            and not await self._is_alert_classification_feature_enabled(watchlist)
        ):
            _logger.warning(
                "Preventing org %s from enabling alert classification on watchlist %s",
                org_id,
                watchlist.id,
            )
            raise HTTPBadRequest(
                reason="Alert Classification is not supported for this watchlist. "
                + "Consult the Developer Network documentation of this "
                + "endpoint for more information"
            )

        async with self.db_pool.rw_transaction() as conn:
            if classifier_id:
                await self._update_watchlist_classifier(
                    conn, watchlist.id, classifier_id
                )
            else:
                # Remove watchlist classifiers in case this used to be a classifier watchlist
                await self._clear_watchlist_classifiers(conn, watchlist.id)

            # Remove the old reports
            # Update reports only if there is a change i.e. if values in request is different from DB
            with self.metrics.timer(
                "watchlists.is_report_update_required.duration"
            ).time():
                update_required = self.is_report_update_required(
                    watchlist.report_ids, db_watchlists[0].report_ids
                )

            if update_required:
                await self._clear_watchlist_reports(
                    conn, watchlist.id, keep_reports=watchlist.report_ids
                )

                await self._add_watchlist_report_ids(
                    conn,
                    org_id=org_id,
                    watchlist_id=watchlist.id,
                    report_ids=watchlist.report_ids,
                )

            await self._update_watchlist(conn, org_id, watchlist, now)

            await self._update_dirty_watchlist_org(conn, org_id=org_id, now=now)

        return (await self.get_watchlists(org_id=org_id, watchlist_id=watchlist.id))[0]

    @staticmethod
    async def _update_watchlist_classifier(conn, watchlist_id: str, classifier_id: int):
        insert_stmt = insert(DbClassifierWatchlist).values(
            watchlist_id=watchlist_id,
            classifier_id=classifier_id,
        )
        upsert_stmt = insert_stmt.on_conflict_do_update(
            index_elements=[DbClassifierWatchlist.watchlist_id],
            set_={"classifier_id": insert_stmt.excluded.classifier_id},
        )

        try:
            await conn.execute(upsert_stmt)
        except IntegrityError as e:
            if not is_fk_violation_exception(e):
                raise e

            _logger.info(
                "Unable to update classifier watchlist.  Watchlist '%s' not found",
                watchlist_id,
            )
            raise WatchlistNotFound(watchlist_id)

    @staticmethod
    async def _clear_watchlist_classifiers(conn, watchlist_id: str) -> None:
        del_stmt = delete(DbClassifierWatchlist).where(
            DbClassifierWatchlist.watchlist_id == watchlist_id
        )
        await conn.execute(del_stmt)

    @staticmethod
    async def _clear_watchlist_reports(
        conn, watchlist_id: str, keep_reports: Iterable[str] = None
    ) -> List[str]:
        del_stmt = delete(DbReportWatchlist).where(
            DbReportWatchlist.watchlist_id == watchlist_id
        )
        if keep_reports:
            del_stmt = del_stmt.where(~DbReportWatchlist.report_id.in_(keep_reports))

        del_stmt = del_stmt.returning(DbReportWatchlist.report_id)
        results = (await conn.execute(del_stmt)).fetchall()

        return [result["report_id"] for result in results]

    @staticmethod
    async def _update_watchlist(
        conn, org_id: str, watchlist: WatchlistV2, now: datetime.datetime
    ) -> None:
        values = {
            "name": watchlist.name,
            "last_update_timestamp": now,
        }
        if watchlist.description is not None:
            values["description"] = watchlist.description
        if watchlist.tags_enabled is not None:
            values["tags_enabled"] = watchlist.tags_enabled
        if watchlist.alerts_enabled is not None:
            values["alerts_enabled"] = watchlist.alerts_enabled
        if watchlist.alert_classification_enabled is not None:
            values[
                "alert_classification_enabled"
            ] = watchlist.alert_classification_enabled

        update_stmt = (
            Watchlist.__table__.update()
            .values(**values)
            .where(Watchlist.org_id == org_id)
            .where(Watchlist.id == watchlist.id)
        )

        try:
            result = await conn.execute(update_stmt)
            if result.rowcount == 0:
                raise WatchlistNotFound(watchlist.id)
        except IntegrityError as e:
            if not is_unique_violation_exception(e):
                raise e

            raise ValueError(f"Watchlist {watchlist.name} already exists")

    async def get_watchlists(
        self,
        org_id: str,
        watchlist_id: str = None,
        require_tags_or_alerts=False,
        use_ro_transaction=False,
    ) -> List[WatchlistV2]:
        watchlists = {}

        join_stmt = (
            Watchlist.__table__.outerjoin(
                DbClassifierWatchlist.__table__,
                DbClassifierWatchlist.watchlist_id == Watchlist.id,
            )
            .outerjoin(
                Classifier.__table__,
                DbClassifierWatchlist.classifier_id == Classifier.id,
            )
            .outerjoin(
                DbReportWatchlist.__table__,
                DbReportWatchlist.watchlist_id == Watchlist.id,
            )
            .outerjoin(
                DbFeed.__table__,
                and_(DbFeed.id == Classifier.value, Classifier.key == "feed_id"),
            )
        )

        sel_stmt = (
            select(
                [
                    Watchlist.id,
                    Watchlist.name,
                    Watchlist.description,
                    Watchlist.tags_enabled,
                    Watchlist.alerts_enabled,
                    Watchlist.alert_classification_enabled,
                    Watchlist.create_timestamp,
                    Watchlist.last_update_timestamp,
                    Classifier.key,
                    Classifier.value,
                    DbReportWatchlist.report_id,
                    DbFeed.alert_classification_feature_enabled,
                ]
            )
            .select_from(join_stmt)
            .where(Watchlist.org_id == org_id)
        )

        if watchlist_id:
            sel_stmt = sel_stmt.where(Watchlist.id == watchlist_id)
        if require_tags_or_alerts:
            sel_stmt = sel_stmt.where(
                or_(
                    Watchlist.tags_enabled == true(), Watchlist.alerts_enabled == true()
                )
            )

        # sort these so it's easier to compare results
        sel_stmt = sel_stmt.order_by(DbReportWatchlist.report_id)

        async with self.db_pool.transaction(read_only=use_ro_transaction) as conn:
            async for result in await conn.stream(sel_stmt):
                watchlist = watchlists.get(result["id"])

                if not watchlist:
                    # gather all the common meta for watchlists
                    #  we will populate report_id and classifer info as we walk the results
                    watchlist = {
                        k: result[k]
                        for k in (
                            "id",
                            "name",
                            "description",
                            "tags_enabled",
                            "alerts_enabled",
                            "create_timestamp",
                            "last_update_timestamp",
                            "alert_classification_enabled",
                        )
                    }
                    watchlists[watchlist["id"]] = watchlist

                if result["key"]:
                    watchlist["classifier"] = {
                        "key": result["key"],
                        "value": result["value"],
                    }

                if result["report_id"]:
                    report_ids = watchlist.get("report_ids")
                    if report_ids is None:
                        report_ids = []
                        watchlist["report_ids"] = report_ids
                    report_ids.append(result["report_id"])

                # Set True/ False on alert_classification_enabled ONLY IF alert classification feature is enabled
                if result["alert_classification_feature_enabled"] is True:
                    if result["alert_classification_enabled"] is None:
                        watchlist["alert_classification_enabled"] = False
                else:
                    watchlist["alert_classification_enabled"] = None

        if watchlist_id and not watchlists:
            raise WatchlistNotFound(watchlist_id)

        # Note if a dict does not have 'classifier' or 'report_ids' then it is assumed to be a report_id watchlist
        #  with all the report_ids removed so we return report_ids=[]
        return [
            WatchlistV2(
                id=v["id"],
                name=v["name"],
                description=v["description"],
                tags_enabled=v["tags_enabled"],
                alerts_enabled=v["alerts_enabled"],
                create_timestamp=v["create_timestamp"],
                last_update_timestamp=v["last_update_timestamp"],
                report_ids=[]
                if "classifier" not in v and "report_ids" not in v
                else v.get("report_ids"),
                classifier=ClassifierKeyValue(
                    key=v["classifier"]["key"], value=v["classifier"]["value"]
                )
                if "classifier" in v
                else None,
                alert_classification_enabled=v["alert_classification_enabled"]
                if "alert_classification_enabled" in v
                else None,
            )
            for v in watchlists.values()
        ]

    async def set_alert_status(
        self, org_id: str, watchlist_id: str, now: datetime.datetime = None
    ) -> bool:
        try:
            watchlists = await self.get_watchlists(org_id, watchlist_id)
            # we assume False as default to prevent watchlists ending up in unhealthy state
            tags = watchlists[0].tags_enabled if watchlists else False
            prevented = (
                True
                if watchlists and await self._check_is_alerting_forbidden(watchlists[0])
                else False
            )

            if not tags:
                _logger.warning(
                    "Preventing org %s from enabling alerts without tags enabled on watchlist %s",
                    org_id,
                    watchlist_id,
                )
                raise InvalidWatchlistAlerts()

            if prevented:
                _logger.warning(
                    "Preventing org %s from enabling alerts on watchlist %s",
                    org_id,
                    watchlist_id,
                )
                raise HTTPBadRequest(
                    reason="Alert generation is not supported for this feed. "
                    + "Consult the Developer Network documentation for this "
                    + "endpoint for more information"
                )

            now = now if now else get_utcnow_with_tzinfo()
            return await self._update_alert_flag(org_id, watchlist_id, True, now)
        except WatchlistNotFound:
            return False  # watchlist not found so couldn't update the alert status

    async def remove_alert_status(
        self, org_id: str, watchlist_id: str, now: datetime.datetime = None
    ) -> bool:
        now = now if now else get_utcnow_with_tzinfo()
        return await self._update_alert_flag(org_id, watchlist_id, False, now)

    async def remove_alert_status_all(
        self, org_id: str, now: datetime.datetime = None
    ) -> bool:
        now = now if now else get_utcnow_with_tzinfo()
        return await self._update_alert_flag(org_id, None, False, now)

    async def get_alert_status(self, org_id: str, watchlist_id: str) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            return await self._get_alert_status(org_id, watchlist_id, conn)

    async def _get_alert_status(self, org_id: str, watchlist_id: str, conn) -> bool:
        query = (
            select([Watchlist.alerts_enabled])
            .where(Watchlist.org_id == org_id)
            .where(Watchlist.id == watchlist_id)
            .limit(1)
        )

        result = (await conn.execute(query)).fetchone()

        return result is not None and result["alerts_enabled"]

    async def get_watchlist_name(self, org_id: str, watchlist_id: str) -> str:
        async with self.db_pool.rw_transaction() as conn:
            return await self._get_watchlist_name(org_id, watchlist_id, conn)

    async def _get_watchlist_name(self, org_id: str, watchlist_id: str, conn) -> str:
        query = (
            select([Watchlist.name])
            .where(Watchlist.org_id == org_id)
            .where(Watchlist.id == watchlist_id)
            .limit(1)
        )

        result = (await conn.execute(query)).fetchone()

        if result is None:
            raise WatchlistNotFound(watchlist_id)

        return result["name"]

    async def _update_alert_flag(
        self, org_id: str, watchlist_id: str, enabled: bool, now: datetime.datetime
    ) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            update_stmt = (
                Watchlist.__table__.update()
                .values(alerts_enabled=enabled, last_update_timestamp=now)
                .where(Watchlist.org_id == org_id)
            )
            # If watchlist_id is passed, modify only that watchlist. Otherwise, we modify all of them for this org.
            if watchlist_id is not None:
                update_stmt = update_stmt.where(Watchlist.id == watchlist_id)
            result = await conn.execute(update_stmt)

            if result.rowcount != 0:
                await self._update_dirty_watchlist_org(conn, org_id=org_id, now=now)
                return True
            else:
                return False

    async def set_tag_status(
        self, org_id: str, watchlist_id: str, now: datetime.datetime = None
    ) -> bool:
        now = now if now else get_utcnow_with_tzinfo()
        return await self._update_tag_flag(org_id, watchlist_id, True, now)

    # Disables document tagging for a watchlist
    async def remove_tag_status(
        self, org_id: str, watchlist_id: str, now: datetime.datetime = None
    ) -> bool:
        now = now if now else get_utcnow_with_tzinfo()
        return await self._update_tag_flag(org_id, watchlist_id, False, now)

    # Disables document tagging for every watchlist in an org
    async def remove_tag_status_all(
        self, org_id: str, now: datetime.datetime = None
    ) -> bool:
        now = now if now else get_utcnow_with_tzinfo()
        return await self._update_tag_flag(org_id, None, False, now)

    async def get_tag_status(self, org_id: str, watchlist_id: str) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            return await self._get_tag_status(org_id, watchlist_id, conn)

    async def _get_tag_status(self, org_id: str, watchlist_id: str, conn) -> bool:
        query = (
            select([Watchlist.tags_enabled])
            .where(Watchlist.org_id == org_id)
            .where(Watchlist.id == watchlist_id)
            .limit(1)
        )

        result = (await conn.execute(query)).fetchone()

        return result is not None and result["tags_enabled"]

    # Enables or disables document tagging for one or all of an org's watchlists
    async def _update_tag_flag(
        self, org_id: str, watchlist_id: str, enabled: bool, now: datetime.datetime
    ) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            update_stmt = (
                Watchlist.__table__.update()
                .values(tags_enabled=enabled, last_update_timestamp=now)
                .where(Watchlist.org_id == org_id)
            )
            # If watchlist_id is passed, modify only that watchlist. Otherwise, we modify all of them for this org.
            if watchlist_id is not None:
                update_stmt = update_stmt.where(Watchlist.id == watchlist_id)
            result = await conn.execute(update_stmt)

            if result.rowcount != 0:
                await self._update_dirty_watchlist_org(conn, org_id=org_id, now=now)
                return True
            else:
                return False

    async def add_ignore_report(self, org_id: str, report_id: str) -> None:
        await self._insert_report_ignores(org_id, [report_id])

    async def add_ignore_report_ioc(
        self, org_id: str, report_id: str, ioc_id: str
    ) -> None:
        await self._insert_report_ioc_ignores(
            org_id, [ReportIOCIgnore(report_id=report_id, ioc_id=ioc_id, ignore=True)]
        )

    async def bulk_update_report_ioc_ignores(
        self, org_id: str, ignores: List[ReportIOCIgnore]
    ) -> List[ReportIOCIgnore]:
        report_updates = defaultdict(list)
        report_ioc_updates = defaultdict(list)
        for ignore in ignores:
            if ignore.ioc_id is None:
                report_updates[ignore.ignore].append(ignore.report_id)
            else:
                report_ioc_updates[ignore.ignore].append(ignore)

        if report_updates[True]:
            await self._insert_report_ignores(org_id, report_ids=report_updates[True])
        if report_updates[False]:
            await self._delete_report_ignores(org_id, report_ids=report_updates[False])
        if report_ioc_updates[True]:
            await self._insert_report_ioc_ignores(org_id, report_ioc_updates[True])
        if report_ioc_updates[False]:
            await self._delete_report_ioc_ignores(org_id, report_ioc_updates[False])

        return ignores

    async def _insert_report_ignores(self, org_id: str, report_ids: List[str]):
        data = [{"org_id": org_id, "report_id": report_id} for report_id in report_ids]

        async with self.db_pool.rw_transaction() as conn:
            insert_stmt = insert(IgnoreReport).values(data)

            upsert_stmt = insert_stmt.on_conflict_do_nothing(
                index_elements=[IgnoreReport.org_id, IgnoreReport.report_id]
            )

            try:
                with self.metrics.timer(
                    "reports.bulk_insert_ignores_table.duration"
                ).time():
                    result = await conn.execute(upsert_stmt)

                if result.rowcount != 0:
                    with self.metrics.timer(
                        "reports.bulk_insert_ignores_update_dirty_watchlists.duration"
                    ).time():
                        await self.report_engine.dirty_org_watchlist_for_reports(
                            conn, report_ids, org_id=org_id
                        )
            except IntegrityError as e:
                if not is_fk_violation_exception(e):
                    raise e

                if len(report_ids) == 1:
                    _logger.info(
                        "Unable to ignore report: %s %s", org_id, report_ids[0]
                    )
                    raise ReportNotFound(report_ids[0])
                else:
                    _logger.info("Unable to ignore one of provided reports: %s", org_id)
                    raise EntityNotFound("One or more reports not found")

    async def _insert_report_ioc_ignores(
        self, org_id: str, ignores: List[ReportIOCIgnore]
    ) -> bool:
        data = [
            {"org_id": org_id, "report_id": i.report_id, "ioc_id": i.ioc_id}
            for i in ignores
        ]

        async with self.db_pool.rw_transaction() as conn:
            insert_stmt = insert(IgnoreReportIOC).values(data)

            upsert_stmt = insert_stmt.on_conflict_do_nothing(
                index_elements=[
                    IgnoreReportIOC.org_id,
                    IgnoreReportIOC.report_id,
                    IgnoreReportIOC.ioc_id,
                ]
            )

            try:
                result = await conn.execute(upsert_stmt)

                if result.rowcount != 0:
                    await self.report_engine.dirty_org_watchlist_for_reports(
                        conn, report_ids=[i.report_id for i in ignores], org_id=org_id
                    )
                    return True
                else:
                    return False
            except IntegrityError as e:
                if not is_fk_violation_exception(e):
                    raise e

                if len(ignores) == 1:
                    _logger.info(
                        "Unable to ignore report ioc: %s %s %s",
                        org_id,
                        ignores[0].report_id,
                        ignores[0].ioc_id,
                    )
                    raise IOCNotFound(ignores[0].report_id, ignores[0].ioc_id)
                else:
                    _logger.info("Unable to ignore report iocs: %s", org_id)
                    raise EntityNotFound("One or more IOCs not found")

    async def remove_ignore_report(self, org_id: str, report_id: str) -> None:
        await self._delete_report_ignores(org_id, report_ids=[report_id])

    async def remove_ignore_report_ioc(
        self, org_id: str, report_id: str, ioc_id: str
    ) -> None:
        await self._delete_report_ioc_ignores(
            org_id,
            ignores=[ReportIOCIgnore(report_id=report_id, ioc_id=ioc_id, ignore=False)],
        )

    async def _delete_report_ignores(
        self, org_id: str, report_ids: Iterable[str]
    ) -> None:
        async with self.db_pool.rw_transaction() as conn:
            del_stmt = (
                delete(IgnoreReport)
                .where(IgnoreReport.org_id == org_id)
                .where(IgnoreReport.report_id.in_(report_ids))
            )

            result = await conn.execute(del_stmt)

            if result.rowcount != 0:
                await self.report_engine.dirty_org_watchlist_for_reports(
                    conn, report_ids=report_ids, org_id=org_id
                )

    async def _delete_report_ioc_ignores(
        self, org_id: str, ignores: Iterable[ReportIOCIgnore]
    ) -> None:
        async with self.db_pool.rw_transaction() as conn:
            del_stmt = (
                delete(IgnoreReportIOC)
                .where(IgnoreReportIOC.org_id == org_id)
                .where(
                    tuple_(IgnoreReportIOC.report_id, IgnoreReportIOC.ioc_id).in_(
                        [(i.report_id, i.ioc_id) for i in ignores]
                    )
                )
            )

            result = await conn.execute(del_stmt)

            if result.rowcount != 0:
                await self.report_engine.dirty_org_watchlist_for_reports(
                    conn, report_ids=[i.report_id for i in ignores], org_id=org_id
                )

    async def get_ignores(
        self, org_id: str, report_ids: Optional[Iterable[str]] = None
    ) -> List[ReportIOCIgnore]:
        ignored_reports = (
            {
                report_id: ReportIOCIgnore(report_id=report_id, ignore=False)
                for report_id in report_ids
            }
            if report_ids
            else {}
        )

        async with self.db_pool.rw_transaction() as conn:
            # Find ignored reports
            sel_stmt = select([IgnoreReport]).where(IgnoreReport.org_id == org_id)
            if report_ids:
                sel_stmt = sel_stmt.where(IgnoreReport.report_id.in_(report_ids))
            async for result in await conn.stream(sel_stmt):
                ignored_reports[result[0].report_id] = ReportIOCIgnore(
                    report_id=result[0].report_id, ignore=True
                )

            ignored = list(ignored_reports.values())

            # Find ignored iocs
            sel_stmt = select([IgnoreReportIOC]).where(IgnoreReportIOC.org_id == org_id)
            if report_ids:
                sel_stmt = sel_stmt.where(IgnoreReportIOC.report_id.in_(report_ids))
            async for result in await conn.stream(sel_stmt):
                ignored.append(
                    ReportIOCIgnore(
                        report_id=result[0].report_id,
                        ioc_id=result[0].ioc_id,
                        ignore=True,
                    )
                )

            return sorted(ignored, key=lambda x: (x.report_id, str(x.ioc_id)))

    async def get_report_ignore(self, org_id: str, report_id: str) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            return (
                await conn.execute(
                    select(
                        exists().where(
                            and_(
                                IgnoreReport.org_id == org_id,
                                IgnoreReport.report_id == report_id,
                            )
                        )
                    )
                )
            ).scalar()

    async def get_report_ioc_ignore(
        self, org_id: str, report_id: str, ioc_id: str
    ) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            return (
                await conn.execute(
                    select(
                        exists().where(
                            and_(
                                IgnoreReportIOC.org_id == org_id,
                                IgnoreReportIOC.report_id == report_id,
                                IgnoreReportIOC.ioc_id == ioc_id,
                            )
                        )
                    )
                )
            ).scalar()

    async def add_report_severity_override(
        self, org_id: str, report_id: str, severity: int
    ) -> bool:
        if severity > 10 or severity < 0:
            raise ValueError(f"Severity ({severity}) must be between 0 and 10")

        async with self.db_pool.rw_transaction() as conn:
            insert_stmt = insert(ReportSeverityOverride).values(
                org_id=org_id, report_id=report_id, severity=severity
            )

            upsert_stmt = insert_stmt.on_conflict_do_update(
                index_elements=[
                    ReportSeverityOverride.org_id,
                    ReportSeverityOverride.report_id,
                ],
                set_={"severity": insert_stmt.excluded.severity},
            )

            try:
                result = await conn.execute(upsert_stmt)

                if result.rowcount == 1:
                    await self.report_engine.dirty_org_watchlist_for_reports(
                        conn, report_ids=[report_id], org_id=org_id
                    )
                    return True
                else:
                    return False
            except IntegrityError as e:
                if not is_fk_violation_exception(e):
                    raise e

                _logger.info("Unable to set report severity: %s %s", org_id, report_id)
                raise ReportNotFound(report_id)

    async def remove_report_severity_override(
        self, org_id: str, report_id: str
    ) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            del_stmt = (
                delete(ReportSeverityOverride)
                .where(ReportSeverityOverride.org_id == org_id)
                .where(ReportSeverityOverride.report_id == report_id)
            )

            result = await conn.execute(del_stmt)

            if result.rowcount != 0:
                await self.report_engine.dirty_org_watchlist_for_reports(
                    conn, report_ids=[report_id], org_id=org_id
                )
                return True
            else:
                return False

    async def get_report_severity_override(
        self, org_id: str, report_id: str
    ) -> Optional[ReportSeverity]:
        async with self.db_pool.rw_transaction() as conn:
            sel_stmt = (
                select(
                    [ReportSeverityOverride.report_id, ReportSeverityOverride.severity]
                )
                .where(ReportSeverityOverride.org_id == org_id)
                .where(ReportSeverityOverride.report_id == report_id)
                .limit(1)
            )

            record = (await conn.execute(sel_stmt)).fetchone()

            return ReportSeverity(report_id, record["severity"]) if record else None

    async def get_report_severity_overrides(self, org_id: str) -> List[ReportSeverity]:
        async with self.db_pool.rw_transaction() as conn:
            sel_stmt = select([ReportSeverityOverride]).where(
                ReportSeverityOverride.org_id == org_id
            )

            return [
                ReportSeverity(
                    report_id=result[0].report_id, severity=result[0].severity
                )
                async for result in await conn.stream(sel_stmt)
            ]

    def _get_aug_report_sel_stmt(self, org_id: str, report_ids: Iterable[str]):
        with self.metrics.timer("feeds.org_augmented.build.watchlist.report").time():
            join_stmt = DbReport.__table__.outerjoin(
                IgnoreReport,
                and_(
                    IgnoreReport.org_id == org_id, IgnoreReport.report_id == DbReport.id
                ),
            )

            sel_stmt = (
                select([DbReport])
                .select_from(join_stmt)
                .where(DbReport.id.in_(report_ids))
                .where(IgnoreReport.report_id == null())
            )

            return sel_stmt

    async def _get_aug_classifier_reports_sel_stmt(
        self, org_id: str, classifier_key: str, classifier_value: str
    ):
        classifier_id = await self.report_engine.get_classifier_id(
            classifier_key, classifier_value
        )

        with self.metrics.timer(
            "feeds.org_augmented.build.watchlist.classifier"
        ).time():
            join_stmt = DbReport.__table__.join(
                ReportClassifier.__table__, ReportClassifier.report_id == DbReport.id
            ).outerjoin(
                IgnoreReport,
                and_(
                    IgnoreReport.org_id == org_id, IgnoreReport.report_id == DbReport.id
                ),
            )

            sel_stmt = (
                select([DbReport])
                .select_from(join_stmt)
                .where(ReportClassifier.classifier_id == classifier_id)
                .where(IgnoreReport.report_id == null())
            )

            return sel_stmt

    async def _get_augmented_watchlist(
        self,
        org_id: str,
        watchlist: WatchlistV2,
        match_type_filter: Dict[str, Callable],
    ) -> Dict[str, AugmentedWatchlist]:
        reports_sel_stmt = None
        if watchlist.report_ids:
            reports_sel_stmt = self._get_aug_report_sel_stmt(
                org_id, watchlist.report_ids
            )
        elif watchlist.classifier:
            reports_sel_stmt = await self._get_aug_classifier_reports_sel_stmt(
                org_id, watchlist.classifier.key, watchlist.classifier.value
            )

        async with self.db_pool.ro_transaction() as conn:
            alert = await self._get_alert_status(org_id, watchlist.id, conn)
            tag = await self._get_tag_status(org_id, watchlist.id, conn)

        match_type_reports = (
            await self._get_augmented_reports(
                org_id, reports_sel_stmt, match_type_filter
            )
            if reports_sel_stmt is not None
            else {}
        )

        augmented_watchlists = {}
        for match_type in match_type_filter:
            reports = match_type_reports.get(match_type, [])

            augmented_watchlists[match_type] = AugmentedWatchlist(
                id=watchlist.id,
                name=watchlist.name,
                action=WatchlistAction(alert, tag),
                reports=reports,
                hash="NA",
                alert_classification_enabled=self.convert_alert_classification_enabled_flag(
                    watchlist.alert_classification_enabled
                ),
            )

        return augmented_watchlists

    async def _get_augmented_reports(
        self, org_id, reports_select, match_type_filters: Dict[str, Callable]
    ) -> Dict[str, List[ApiReport]]:
        reports = defaultdict(list)

        async with self.db_pool.ro_transaction() as conn:
            # build report severity override dictionary
            #  Todo - is the result set too broad (org_id) not that we can't limit by feed
            severities = {
                severity.report_id: severity.severity
                for severity in await self.get_report_severity_overrides(org_id)
            }

            ioc_ignore_select = (
                select([IgnoreReportIOC])
                .where(IgnoreReportIOC.report_id == IOC.report_id)
                .where(IgnoreReportIOC.ioc_id == IOC.id)
                .where(IgnoreReportIOC.org_id == org_id)
            )

            report_results = {}
            async for result in await conn.stream(reports_select):
                report_results[result[0].id] = result[0]

            report_iocs = defaultdict(list)
            for chunk in chunker(250, report_results.keys()):
                async for result in await conn.stream(
                    select([IOC])
                    .where(IOC.report_id.in_(chunk))
                    .where(~exists(ioc_ignore_select))
                ):
                    report_iocs[result[0].report_id].append(result[0])

            for report in report_results.values():
                iocs_v2 = defaultdict(list)
                for iocs_v2_result in report_iocs[report.id]:
                    ioc_v2 = ReportEngine.iocs_v2_to_api(iocs_v2_result)

                    for match_type, filter_func in match_type_filters.items():
                        if filter_func(ioc_v2):
                            iocs_v2[match_type].append(ioc_v2)
                            break

                # update severity if it was overridden
                severity = severities.get(report.id, report.severity)

                for match_type, iocs in iocs_v2.items():
                    if iocs:
                        reports[match_type].append(
                            self._report_to_api(
                                report, iocs, len(iocs), severity=severity
                            )
                        )

        return reports

    async def get_org_augmented_watchlists_by_match_type(
        self, org_id: str, match_type_filters: Dict[str, Callable]
    ) -> Dict[str, AugmentedWatchlists]:
        with self.metrics.timer("feeds.org_augmented.build.all").time():
            watchlists_by_type = defaultdict(list)

            for wl in await self.get_watchlists(
                org_id, require_tags_or_alerts=True, use_ro_transaction=True
            ):
                filtered = await self._get_augmented_watchlist(
                    org_id, wl, match_type_filters
                )

                for match_type, watchlist in filtered.items():
                    watchlists_by_type[match_type].append(watchlist)

            augmented_watchlists = {}

            for match_type in match_type_filters:
                watchlists = watchlists_by_type[match_type]

                watchlist_meta = {}
                reports = {}
                report_watchlists = defaultdict(list)

                for watchlist in sorted(watchlists, key=lambda x: x.id):
                    if watchlist.id not in watchlist_meta:
                        watchlist_meta[watchlist.id] = WatchlistMeta(
                            id=watchlist.id,
                            name=watchlist.name,
                            alert=watchlist.action.alert,
                            tag=watchlist.action.tag,
                            alert_classification_enabled=self.convert_alert_classification_enabled_flag(
                                watchlist.alert_classification_enabled
                            ),
                        )

                    for report in watchlist.reports:
                        report_watchlists[report.id].append(watchlist.id)
                        if report.id not in reports:
                            reports[report.id] = report

                augmented_reports = []
                for report in reports.values():
                    augmented_reports.append(
                        AugmentedReport(
                            id=report.id,
                            watchlists=report_watchlists[report.id],
                            timestamp=datetime_to_sec_timestamp(report.timestamp),
                            link=report.link,
                            title=report.title,
                            description=report.description,
                            severity=report.severity,
                            tags=report.tags,
                            iocs=report.iocs_v2,
                            visibility=report.visibility,
                        )
                    )

                augmented_watchlists[match_type] = AugmentedWatchlists(
                    watchlists=list(watchlist_meta.values())
                    if augmented_reports
                    else [],
                    reports=augmented_reports,
                    hash="NA",
                )

                # try to play nice with other tasks
                await asyncio.sleep(0)

            return augmented_watchlists

    async def add_report(
        self, org_id: str, report: ApiReport, access: str = ReportEngine.PRIVATE_LABEL
    ) -> ApiReport:
        classifiers = [
            {"key": ReportEngine.SOURCE_CLASSIFIER, "value": org_id},
            {"key": ReportEngine.ACCESS_CLASSIFIER, "value": access},
        ]

        classifier_ids = await self.report_engine.insert_classifiers(classifiers)

        await self.report_engine.validate_reports(org_id, [report])

        async with self.db_pool.rw_transaction() as conn:
            await self.report_engine.add_report(
                conn, report=report, classifier_ids=classifier_ids
            )

        await self.publisher.publish_report_notifications([report.id])

        return report

    async def update_report(self, org_id: str, report: ApiReport) -> ApiReport:
        await self.report_engine.verify_report_permissions(org_id, report.id)

        await self.report_engine.validate_reports(org_id, [report])

        async with self.db_pool.rw_transaction() as conn:
            await self.report_engine.upsert_report(conn, report)

        await self.publisher.publish_report_notifications([report.id])
        return report

    async def get_report(
        self, org_id: str, report_id: str, iocs_start_row=None, iocs_row_count=None
    ) -> ApiReport:
        return await self.report_engine.get_report(
            org_id, report_id, iocs_start_row, iocs_row_count
        )

    async def get_report_title(self, report_id: str) -> str:
        return await self.report_engine.get_report_title(report_id)

    async def get_extended_reports(
        self, report_ids: Iterable[str]
    ) -> Dict[str, ExtendedReport]:
        reports = await self.report_engine.get_reports_from_ids(report_ids)
        found = {report.id for report in reports}

        report_sources = await self.report_engine.get_report_classifier_values(
            found, ReportEngine.SOURCE_CLASSIFIER
        )
        report_accesses = await self.report_engine.get_report_classifier_values(
            found, ReportEngine.ACCESS_CLASSIFIER
        )

        extended_reports = {}
        for report in reports:
            report_source = report_sources.get(report.id)
            if not report_source:
                _logger.info(
                    "get_extended_reports: Missing report source for report %s",
                    report.id,
                )
                continue

            report_access = report_accesses.get(report.id)
            if not report_access:
                _logger.info(
                    "get_extended_reports: Missing report access for report %s",
                    report.id,
                )
                continue

            extended_reports[report.id] = ExtendedReport.from_basic(
                report, source=report_source, access=report_access, feedinfo=None
            )

        return extended_reports

    async def delete_report(self, org_id: str, report_id) -> bool:
        await self.report_engine.verify_report_permissions(org_id, report_id)

        async with self.db_pool.rw_transaction() as conn:
            result = await self.report_engine.delete_report(conn, report_id)

        if result:
            await self.publisher.publish_report_notifications([report_id])
            return True
        else:
            return False

    async def add_report_iocs(
        self,
        org_id: str,
        report_id: str,
        iocs: Iterable[IOC_V2],
        report_timestamp: datetime.datetime,
    ) -> bool:
        await self.report_engine.verify_report_permissions(
            org_id=org_id, report_id=report_id
        )

        async with self.db_pool.rw_transaction() as conn:
            success = await self.report_engine.add_report_iocs(
                conn, report_id, iocs, report_timestamp
            )
            if success:
                await self.report_engine.dirty_org_watchlist_for_reports(
                    conn, [report_id]
                )

        # Do publish step outside of the sql transaction
        if success:
            await self.publisher.publish_report_notifications([report_id])

        return success

    async def delete_report_iocs(
        self,
        org_id: str,
        report_id: str,
        ioc_ids: Iterable[str],
        report_timestamp: datetime.datetime,
    ) -> bool:
        await self.report_engine.verify_report_permissions(
            org_id=org_id, report_id=report_id
        )

        async with self.db_pool.rw_transaction() as conn:
            success = await self.report_engine.delete_report_iocs(
                conn, report_id, ioc_ids, report_timestamp
            )
            if success:
                await self.report_engine.dirty_org_watchlist_for_reports(
                    conn, [report_id]
                )

        # Do publish step outside of the sql transaction
        if success:
            await self.publisher.publish_report_notifications([report_id])

        return success

    async def get_dirty_org_for_build(
        self, now: datetime.datetime
    ) -> Optional[DirtyOrgAugmentationTuple]:
        dirty_threshold = now - datetime.timedelta(
            seconds=self.config.getint("watchlist", "dirty_threshold_sec", fallback=30)
        )

        sel_stmt = (
            select(
                [
                    OrganizationWatchlistState.org_id,
                    OrganizationWatchlistState.dirty_timestamp,
                    func.now().label("timestamp"),
                ]
            )
            .where(OrganizationWatchlistState.dirty_timestamp < dirty_threshold)
            .with_for_update(skip_locked=True)
            .order_by(OrganizationWatchlistState.dirty_timestamp.asc())
            .limit(1)
        )

        del_stmt = delete(OrganizationWatchlistState).returning(
            OrganizationWatchlistState.org_id,
            OrganizationWatchlistState.dirty_timestamp,
            func.now().label("timestamp"),
        )

        try:
            async with self.db_pool.rw_transaction() as conn:
                result = (await conn.execute(sel_stmt)).fetchone()

                if not result:
                    return None
                else:
                    _logger.info(f"Found dirty org {result}")
                    dirty = DirtyOrgAugmentationTuple(
                        result["org_id"], result["dirty_timestamp"], result["timestamp"]
                    )

                    del_stmt = del_stmt.where(
                        OrganizationWatchlistState.org_id == result["org_id"]
                    ).execution_options(synchronize_session=False)
                    result = (await conn.execute(del_stmt)).fetchone()
                    _logger.info(f"Removed dirty org {result}")

                    return dirty
        except Exception as e:
            _logger.warning("Error getting dirty watchlist org: %s", str(e))
            return None

    async def set_org_watchlist_buildtime(
        self, org_id: str, timestamp: datetime.datetime
    ) -> None:
        insert_stmt = insert(OrganizationWatchlistBuild).values(
            org_id=org_id, build_timestamp=timestamp
        )

        upsert_stmt = insert_stmt.on_conflict_do_update(
            index_elements=[OrganizationWatchlistBuild.org_id],
            set_={"build_timestamp": insert_stmt.excluded.build_timestamp},
        )

        async with self.db_pool.rw_transaction() as conn:
            await conn.execute(upsert_stmt)

    async def get_stale_watchlist_orgs(self, threshold: datetime.datetime) -> List[str]:
        sel_stmt = select([OrganizationWatchlistBuild.org_id]).where(
            OrganizationWatchlistBuild.build_timestamp < threshold
        )

        org_ids = []
        async with self.db_pool.rw_transaction() as conn:
            async for result in await conn.stream(sel_stmt):
                org_ids.append(result["org_id"])

        return org_ids

    async def publish_watchlists(
        self,
        org_id: str,
        timestamp: datetime.datetime,
        match_type: str,
        watchlists: AugmentedWatchlists,
    ) -> None:
        s3_bucket = self.config.get("watchlist", "augmentation_bucket")
        key = f"watchlists/org={org_id}/watchlist.{match_type}.json"
        # Encode the watchlists into a binary file and MD5 it. We'll compare that against
        # what's already in S3 and if they're different, we'll update S3 with what we have here.
        new_watchlists_bin = BytesIO(json_dumps(watchlists).encode())
        new_watchlists_md5 = md5(
            new_watchlists_bin.getvalue(), usedforsecurity=False
        ).hexdigest()
        new_watchlists_meta = {"md5sum": new_watchlists_md5}

        # Retrieve the existing watchlists file's metadata (if any)
        async with aws_client(self.config, "s3") as s3:
            try:
                old_watchlists_info = await s3.head_object(Bucket=s3_bucket, Key=key)
                if (
                    "Metadata" in old_watchlists_info
                    and "md5sum" in old_watchlists_info["Metadata"]
                ):
                    # Compare the MD5s of previous and new watchlist lists. Only upload if they're not equal
                    old_watchlists_md5 = old_watchlists_info["Metadata"]["md5sum"]
                    if old_watchlists_md5 != new_watchlists_md5:
                        _logger.info(
                            "Publishing watchlist s3://%s/%s (created: %s) (md5 old:%s new:%s) ",
                            s3_bucket,
                            key,
                            timestamp,
                            old_watchlists_md5,
                            new_watchlists_md5,
                        )
                        await self._upload_watchlists(
                            s3,
                            s3_bucket,
                            key,
                            new_watchlists_bin,
                            timestamp,
                            new_watchlists_meta,
                        )
                    # Watchlist didn't change and shouldn't be considered dirty
                    else:
                        await OrgWatchlistStateUpdater(self.config).delete_s3_dirty_tag(
                            org_id, match_type, s3
                        )
                # If the MD5 isn't in the metadata, we can't reason about it, so we have to upload.
                else:
                    _logger.info(
                        "Publishing watchlist s3://%s/%s (created: %s) (md5 old:%s new:%s) ",
                        s3_bucket,
                        key,
                        timestamp,
                        "N/A",
                        new_watchlists_md5,
                    )
                    await self._upload_watchlists(
                        s3,
                        s3_bucket,
                        key,
                        new_watchlists_bin,
                        timestamp,
                        new_watchlists_meta,
                    )
            except ClientError as e:
                status = e.response["ResponseMetadata"]["HTTPStatusCode"]
                if status == 404:
                    _logger.info(
                        "Need to publish watchlists for new org s3://%s/%s.",
                        s3_bucket,
                        key,
                    )
                    # If there was no existing file listing watchlists, we upload
                    await self._upload_watchlists(
                        s3,
                        s3_bucket,
                        key,
                        new_watchlists_bin,
                        timestamp,
                        new_watchlists_meta,
                    )
                else:
                    _logger.error(
                        "Unhandled error getting org watchlists file (s3://%s/%s), %s",
                        s3_bucket,
                        key,
                        e,
                    )

    @staticmethod
    async def _upload_watchlists(
        s3_client, bucket, key, watchlists_binary, timestamp, object_metadata
    ) -> None:
        await s3_client.upload_fileobj(
            watchlists_binary,
            Bucket=bucket,
            Key=key,
            ExtraArgs={
                "Tagging": f"created={datetime_to_ms_timestamp(timestamp)}",
                "Metadata": object_metadata,
            },
        )

    async def get_watchlist_telemetry(
        self, org_id: str, watchlist_ids: Iterable[str], intervals: Iterable[int]
    ) -> List[WatchlistTelemetry]:

        validate_owner_tasks = [
            self.validate_watchlist_owner(org_id, watchlist_id)
            for watchlist_id in watchlist_ids
        ]
        owner_results = await asyncio.gather(*validate_owner_tasks)
        for watchlist_id, valid in itertools.zip_longest(watchlist_ids, owner_results):
            if not valid:
                raise WatchlistNotFound(watchlist_id=watchlist_id)

        telemetry_tasks = [
            self.telemetry_client.get_watchlist_telemetry(watchlist_id, intervals)
            for watchlist_id in watchlist_ids
        ]
        telemetry_results = await asyncio.gather(*telemetry_tasks)

        # turn list of lists of WatchlistTelemetry objects into single list of WatchlistTelemetry objects
        return [telemetry for results in telemetry_results for telemetry in results]

    async def validate_watchlist_owner(self, org_id: str, watchlist_id: str) -> bool:
        async with self.db_pool.rw_transaction() as conn:
            sel_stmt = (
                select([Watchlist.id])
                .where(Watchlist.org_id == org_id)
                .where(Watchlist.id == watchlist_id)
                .limit(1)
            )

            result = (await conn.execute(sel_stmt)).fetchone()

            return result is not None

    @staticmethod
    def _report_to_api(
        record: DbReport,
        iocs_v2: Iterable[IOC_V2] = None,
        iocs_total_count: int = None,
        severity: int = None,
    ) -> ApiReport:
        return ApiReport(
            id=record.id,
            timestamp=record.timestamp.replace(tzinfo=datetime.timezone.utc),
            link=record.link,
            title=record.title,
            severity=severity if severity is not None else record.severity,
            tags=json.loads(record.tags) if record.tags else None,
            iocs=None,
            iocs_v2=iocs_v2 if iocs_v2 else None,
            iocs_total_count=iocs_total_count if iocs_total_count else 0,
            description=record.description,
            visibility=record.visibility,
        )

    @staticmethod
    def remove_empty_alert_classification_field_v2(watchlist):
        if watchlist.alert_classification_enabled is None:
            new_watchlist = watchlist._asdict()
            del new_watchlist["alert_classification_enabled"]
            return namedtuple("WatchlistV2", new_watchlist.keys())(
                *new_watchlist.values()
            )

        return watchlist

    @staticmethod
    def remove_empty_alert_classification_field(watchlist):
        if watchlist is not None:
            new_watchlist = watchlist._asdict()
            classifier = new_watchlist["classifier"]
            report = new_watchlist["report"]
            if classifier and classifier.alert_classification_enabled is None:
                new_classifier = classifier._asdict()
                del new_classifier["alert_classification_enabled"]
                new_watchlist["classifier"] = namedtuple(
                    "ClassifierWatchlist", new_classifier.keys()
                )(*new_classifier.values())

            if report and report.alert_classification_enabled is None:
                new_report = report._asdict()
                del new_report["alert_classification_enabled"]
                new_watchlist["report"] = namedtuple(
                    "ReportWatchlist", new_report.keys()
                )(*new_report.values())

            return namedtuple("Watchlist", new_watchlist.keys())(
                *new_watchlist.values()
            )

        return watchlist

    @staticmethod
    def convert_alert_classification_enabled_flag(flag) -> bool:
        if flag is None:
            return False

        return flag

    # A utility method to update Org Watchlist state table for testing
    async def _init_watchlist_state_for_testing(self):
        async with self.db_pool.rw_transaction() as conn:
            sel_stmt = select([Watchlist.org_id]).distinct(Watchlist.org_id)

            results = (await conn.execute(sel_stmt)).fetchall()

            for result in results:
                org_id = result["org_id"]

                insert_stmt = (
                    insert(OrganizationWatchlistState)
                    .values(org_id=org_id, dirty_timestamp=EPOCH)
                    .returning(OrganizationWatchlistState.org_id)
                )

                upsert_stmt = insert_stmt.on_conflict_do_update(
                    index_elements=[OrganizationWatchlistState.org_id],
                    set_={"dirty_timestamp": insert_stmt.excluded.dirty_timestamp},
                    where=OrganizationWatchlistState.dirty_timestamp == null(),
                )

                await conn.execute(upsert_stmt)

    def is_report_update_required(
        self, in_report_ids: List[str], db_report_ids: List[str]
    ) -> bool:
        if in_report_ids is None and db_report_ids is not None:
            return True

        if in_report_ids is not None and db_report_ids is None:
            return True

        if in_report_ids is not None:
            in_report_ids.sort()

        if db_report_ids is not None:
            db_report_ids.sort()

        if in_report_ids == db_report_ids:
            _logger.info(
                "Report Ids in request is same as in DB. Skipping reports update"
            )
            return False

        return True

    async def get_mdr_reports(self, org_id):
        mdr_reports = []
        mdr_feeds = await self._get_mdr_feeds(org_id)

        for mdr_feed_id in mdr_feeds:
            feed_classifier_id = await self.report_engine.get_classifier_id(
                FeedEngine.FEED_ID_CLASSIFIER, mdr_feed_id
            )
            reports = await self.report_engine.get_reports_with_classifiers(
                [feed_classifier_id]
            )
            for report in reports:
                mdr_reports.append(report.id)

        return mdr_reports

    async def _get_mdr_feeds(self, org_id):
        scope = '["mdforeedr"]'
        mdr_feeds = []
        sel_stmt = FeedEngine.feedinfos_select(
            org_id=org_id, include_public=True
        ).where(DbFeed.scope == scope)
        async with self.db_pool.ro_transaction() as conn:
            async for result in await conn.stream(sel_stmt):
                mdr_feeds.append(result[0].id)

        return mdr_feeds
