import asyncio

import logging
from uuid import uuid4
from typing import Iterable

from pscr_common.utils import chunker

from feed_manager.utils import json_dumps, aws_client, lazy_join

_logger = logging.getLogger(__name__)


class Publisher(object):
    OP_PARTIAL = "partial"
    OP_PARTIAL_COMPLETE = "partial-complete"

    def __init__(
        self,
        config,
        db_pool,
        stream_name,
    ):
        self.config = config
        self.db_pool = db_pool
        self.stream_name = stream_name
        self.max_kinesis_record_report_ids = int(
            config.get("feedsearch", "max_kinesis_record_report_ids")
        )
        self.max_kinesis_put_records_count = int(
            config.get("feedsearch", "max_kinesis_put_records_count")
        )

    async def publish_report_notifications(self, report_ids: Iterable[str]):
        _logger.debug("Publishing report notifications: %s", lazy_join(",", report_ids))

        for records in chunker(
            self.max_kinesis_put_records_count, self._get_kinesis_records(report_ids)
        ):
            await self._put_records(records)

    def _get_kinesis_records(self, report_ids):
        for report_id_chunk in chunker(self.max_kinesis_record_report_ids, report_ids):
            data = json_dumps({"report_ids": report_id_chunk})
            yield {"Data": data, "PartitionKey": uuid4().hex}

    async def _put_records(self, records):
        async with aws_client(self.config, "kinesis") as kinesis:
            records_to_send = records
            while records_to_send:
                _logger.debug("Publisher put_records: %s", len(records_to_send))
                response = await kinesis.put_records(
                    Records=records_to_send, StreamName=self.stream_name
                )

                if response["FailedRecordCount"]:
                    records_to_send = [
                        records_to_send[i]
                        for i, record in enumerate(response["Records"])
                        if "ErrorCode" in record
                    ]

                    _logger.debug("Resending %s records", len(records_to_send))
                    await asyncio.sleep(0.5)
                else:
                    records_to_send = []
