import logging
import rapidjson
from aiohttp import web

from feed_manager.api.models.feed_models import FeedInfo, Report
from feed_manager.engine.feed_engine import FeedEngine
from feed_manager.web.route_helpers import get_request_org, make_response_json
from feed_manager.web.auth import Resource
from pscr_common.web.auth import AuthRouteTableDef, Action

_logger = logging.getLogger(__name__)


def build_route_table(engine: FeedEngine) -> AuthRouteTableDef:
    routes = AuthRouteTableDef()

    @routes.get("/healthcheck")
    async def healthcheck(request: web.Request):
        request.do_not_log = True

        return make_response_json()

    @routes.post("/v1/feed", Resource.feeds, [Action.create])
    async def create_feed(request: web.Request):
        params = await request.json(loads=rapidjson.loads)

        org_id = get_request_org(request)

        params["feedinfo"]["owner"] = org_id

        reports = [Report.from_dict(report) for report in params["reports"]]

        feedinfo = await engine.create_feed(
            org_id, FeedInfo.from_dict(params["feedinfo"]), reports
        )

        return make_response_json({"feed_id": feedinfo.id})

    @routes.get("/v1/feed", Resource.feeds, [Action.read])
    async def get_feeds(request: web.Request):
        include_public = request.query.get("include_public", "false").lower() == "true"

        feeds = await engine.get_feeds(get_request_org(request), include_public)

        return make_response_json({"feeds": feeds})

    @routes.get("/v1/feed/{feed_id}", Resource.feeds, [Action.read])
    async def get_feed(request: web.Request):
        feed_id = request.match_info["feed_id"]
        feed = await engine.get_feed(get_request_org(request), feed_id)

        return make_response_json({"feed": feed})

    @routes.delete("/v1/feed/{feed_id}", Resource.feeds, [Action.delete])
    async def remove_feed(request: web.Request):
        feed_id = request.match_info["feed_id"]

        await engine.remove_feed(get_request_org(request), feed_id)

        return make_response_json({"success": True})

    @routes.get("/v1/feed/{feed_id}/feedinfo", Resource.feeds, [Action.read])
    async def get_feedinfo(request: web.Request):
        feed_id = request.match_info["feed_id"]

        feedinfo = await engine.get_feedinfo(get_request_org(request), feed_id)

        return make_response_json({"feedinfo": feedinfo})

    @routes.put("/v1/feed/{feed_id}/feedinfo", Resource.feeds, [Action.update])
    async def update_feedinfo(request: web.Request):
        feed_id = request.match_info["feed_id"]
        params = await request.json(loads=rapidjson.loads)

        org_id = get_request_org(request)

        if "owner" not in params:
            params["owner"] = org_id

        params["id"] = feed_id

        await engine.upsert_feedinfo(org_id, feed_id, FeedInfo.from_dict(params))

        return make_response_json({"success": True})

    @routes.get("/v1/feed/{feed_id}/report", Resource.feeds, [Action.read])
    async def get_reports(request: web.Request):
        feed_id = request.match_info["feed_id"]

        reports = await engine.get_feed_reports(get_request_org(request), feed_id)

        return make_response_json({"reports": reports})

    @routes.get("/v1/feed/{feed_id}/report/{report_id}", Resource.feeds, [Action.read])
    async def get_report(request: web.Request):
        feed_id = request.match_info["feed_id"]
        report_id = request.match_info["report_id"]

        report = await engine.get_feed_report(
            get_request_org(request), feed_id, report_id
        )

        return make_response_json({"report": report})

    @routes.post("/v1/feed/{feed_id}/report", Resource.feeds, [Action.create])
    async def replace_reports(request: web.Request):
        feed_id = request.match_info["feed_id"]
        params = await request.json(loads=rapidjson.loads)

        org_id = get_request_org(request)

        reports = [Report.from_dict(report) for report in params["reports"]]

        inserted, updated, deleted = await engine.replace_reports(
            org_id, feed_id, reports
        )

        return make_response_json({"success": inserted + updated + deleted > 0})

    @routes.put(
        "/v1/feed/{feed_id}/report/{report_id}", Resource.feeds, [Action.update]
    )
    async def update_report(request: web.Request):
        feed_id = request.match_info["feed_id"]
        report_id = request.match_info["report_id"]

        params = await request.json(loads=rapidjson.loads)

        org_id = get_request_org(request)

        params["id"] = report_id

        await engine.upsert_report(org_id, feed_id, Report.from_dict(params))

        return make_response_json({"success": True})

    @routes.delete(
        "/v1/feed/{feed_id}/report/{report_id}", Resource.feeds, [Action.delete]
    )
    async def remove_report(request: web.Request):
        feed_id = request.match_info["feed_id"]
        report_id = request.match_info["report_id"]

        org_id = get_request_org(request)

        await engine.delete_report(org_id, feed_id, report_id)

        return make_response_json({"success": True})

    return routes
